from peegy.definitions.channel_definitions import Domain
from peegy.tools.units.unit_tools import set_default_unit
import pandas as pd
import astropy.units as u


class EegPeak(object):
    def __init__(self,
                 channel: str = None,
                 x: float = None,
                 rn: float = None,
                 amp: float = None,
                 amp_snr: float = None,
                 significant: bool = False,
                 peak_label: str = None,
                 show_label: bool = True,
                 show_peak: bool = True,
                 positive: bool = True,
                 domain: Domain = Domain.time,
                 spectral_phase: float = None):
        self.channel = channel
        self.x = x
        self.rn = rn
        self.amp = amp
        self.amp_snr = amp_snr
        self.significant = significant
        self.peak_label = peak_label
        self.show_label = show_label
        self.show_peak = show_peak
        self.positive = positive
        self.domain = domain
        self.spectral_phase = spectral_phase


class TimePeakWindow(object):
    def __init__(self,
                 label: str = None,
                 ini_time: u.quantity.Quantity = None,
                 end_time: u.quantity.Quantity = None,
                 ini_ref: str = None,
                 end_ref: str = None,
                 require_ini_ref: bool = False,
                 require_end_ref: bool = False,
                 global_ref: str = None,
                 require_global_ref: bool = False,
                 ini_global_ref: str = None,
                 end_global_ref: str = None,
                 positive_peak: bool = False,
                 show_label: bool = True,
                 show_peak: bool = True,
                 force_search: bool = True,
                 channel: str = None,
                 target_channels: [str] = None,
                 exclude_channels: [str] = None):
        self._label = label
        self._ini_time = set_default_unit(ini_time, u.s)
        self._end_time = set_default_unit(end_time, u.s)
        self.ini_ref = ini_ref
        self.end_ref = end_ref
        self.require_ini_ref = require_ini_ref
        self.require_end_ref = require_end_ref
        self.global_ref = global_ref
        self.require_global_ref = require_global_ref
        self.ini_global_ref = ini_global_ref
        self.end_global_ref = end_global_ref
        self.positive_peak = positive_peak
        self.show_label = show_label
        self.show_peak = show_peak
        self.force_search = force_search
        self.channel = channel
        self.target_channels = target_channels
        self.exclude_channels = exclude_channels

    def get_ini_time(self):
        return self._ini_time

    def set_ini_time(self, value):
        self._ini_time = value

    ini_time = property(get_ini_time, set_ini_time)

    def get_end_time(self):
        _end_time = self._end_time
        if self._end_time is None:
            _end_time = self._ini_time
        return _end_time

    def set_end_time(self, value):
        self._end_time = value
    end_time = property(get_end_time, set_end_time)

    def set_label(self, value):
        self._label = value

    def get_label(self):
        _label = self._label
        if self._label is None:
            _label = '{:.2e}'.format(self.ini_time)
        return _label
    label = property(get_label, set_label)


class PeaksContainer:
    def __init__(self, channel_label: str = None):
        self.channel_label = channel_label
        self._peaks = []

    def get_peaks(self):
        return self._peaks

    def append(self, value: EegPeak = None):
        self._peaks.append(value)

    peaks = property(get_peaks)

    def to_pandas(self):
        _data_pd = pd.DataFrame([_peak.__dict__ for _peak in self._peaks])
        return _data_pd


class PeakToPeakMeasure(object):
    def __init__(self,
                 ini_peak: str = None,
                 end_peak: str = None):
        """
        Class container. This class defines the peaks labels that will be used to measure peak-to-peak amplitudes.
        :param ini_peak: the label of the initial peak
        :param end_peak: the label of the end peak
        """
        self.ini_peak = ini_peak
        self.end_peak = end_peak


class PeakToPeakQuantity(object):
    def __init__(self,
                 channel: str = None,
                 ini_time: u.Quantity = None,
                 end_time: u.Quantity = None,
                 max_time: u.Quantity = None,
                 min_time: u.Quantity = None,
                 amp: u.Quantity = None,
                 rn: u.Quantity = None,
                 amp_snr: u.Quantity = None,
                 amp_label: str = None
                 ):
        """
        Class container. This class defines the parameters that will be used to store peak-to-peak measurements.
        :param channel:
        :param ini_time: time of initial peak in a peak-to-peak measurement
        :param end_time: time of end peak in a peak-to-peak measurement
        :param max_time: time of the maximum peak
        :param min_time: time of the minimum peak
        :param amp: peak-to-peak amplitude
        :param rn: estimated residual noise in peak-to-peak measurement
        :param amp_snr: estimated peak-to-peak amplitude signal-to-noise ratio
        :param amp_label: label to identify this measurement
        """
        self.channel = channel
        self.ini_time = ini_time
        self.max_time = max_time
        self.min_time = min_time
        self.amp = amp
        self.rn = rn
        self.amp_snr = amp_snr
        self.amp_label = amp_label


class PeakToPeakAmplitudeContainer:
    def __init__(self,
                 channel_label: str = None):
        self.channel_label = channel_label
        self._peaks = []

    def get_peaks(self):
        return self._peaks

    def append(self, value: PeakToPeakQuantity = None):
        self._peaks.append(value)

    peaks = property(get_peaks)

    def to_pandas(self):
        _data_pd = pd.DataFrame([_peak.__dict__ for _peak in self._peaks])
        return _data_pd


class TimeROI(object):
    """
    Object used to define the timepoints in a region of interest (ROI), for time-window based analysis. The required
    parameters  are the beginning (initial) and ending timepoints in seconds. To distinguish different ROI windows,
    each objects can  be linked to a measure (string), a label (string), a condition (string), an (EEG) channel (str),
    or a value (float). This class will return a  numpy array. The TimeROI object is used by classes like AverageEpochs
    or HotellingT2Test.
    Example to define a ROI window for SNR estimation starting at 150 ms and ending at 350 ms:
    roi_windows = np.array([TimeROI(ini_time=150.0e-3, end_time=350.0e-3, measure="snr", label="itd_snr")])
    """
    def __init__(self,
                 ini_time: u.quantity.Quantity = None,
                 end_time: u.quantity.Quantity = None,
                 measure: str = None,
                 label: str = None,
                 condition: str = None,
                 value: u.quantity.Quantity = None,
                 show_window: bool = True,
                 show_label: bool = False,
                 ):
        """

        :param ini_time: initial time of the window (in seconds)
        :param end_time: end time of the window (in seconds)
        :param measure: str measure to be performed e.g. "snr", "rms" (this key value can be used to trigger a given
        action within a function requiring a ROI window
        :param label: str defining a desired label for this ROI
        :param condition: str defining a desired condition
        :param value: output value to store the output  of a given measure
        :param show_window: boolean use when plotting data. If true, time interval will be shown in plots
        :param show_label: boolean use when plotting data. If true, label will be shown in plots
        """
        self.ini_time = set_default_unit(ini_time, u.s)
        self.end_time = set_default_unit(end_time, u.s)
        self.measure = measure
        self.label = label
        self.condition = condition
        self.value = value
        self.show_window = show_window
        self.show_label = show_label


class Marker(object):
    def __init__(self,
                 x_ini: u.Quantity = 0,
                 x_end: u.Quantity = 0,
                 y_ini: u.Quantity = 0,
                 y_end: u.Quantity = 0,
                 label: str = None,
                 channel: str = None):
        self.x_ini = x_ini
        self.x_end = x_end
        self.y_ini = y_ini
        self.y_end = y_end
        self.channel = channel
        self.label = label



from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.epochs_processing_tools import et_mean, et_average_frequency_transformation,\
    et_average_time_frequency_transformation
import multiprocessing
from peegy.plot import eeg_ave_epochs_plot_tools as eegpt
from peegy.definitions.channel_definitions import Domain
import pyfftw
import numpy as np
import os
import astropy.units as u
from PyQt5.QtCore import QLibraryInfo
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class HilbertEnvelope(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess, high_pass=None, low_pass=None, **kwargs):
        """
        This process computes the  Hilbert Envelope of EEG data
        :param input_process: InputOutputProcess Class
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(HilbertEnvelope, self).__init__(input_process=input_process, **kwargs)

    def transform_data(self):
        data = self.input_node.data.copy()
        _fft = pyfftw.builders.fft(data, overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                   threads=multiprocessing.cpu_count())
        fx = _fft()
        n = fx.shape[0]
        h = np.zeros(n)
        if n % 2 == 0:
            h[0] = h[n // 2] = 1
            h[1:n // 2] = 2
        else:
            h[0] = 1
            h[1:(n + 1) // 2] = 2
        _ifft = pyfftw.builders.ifft(fx * h.reshape(-1, 1), overwrite_input=False, planner_effort='FFTW_ESTIMATE',
                                     axis=0,
                                     threads=multiprocessing.cpu_count())
        hilbert_data = _ifft()
        self.output_node.data = np.abs(hilbert_data)


class InducedResponse(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 weighted_average=True,
                 n_tracked_points=256,
                 block_size=5,
                 roi_windows=None, **kwargs):
        super(InducedResponse, self).__init__(input_process=input_process, **kwargs)
        self.weighted_average = weighted_average
        self.n_tracked_points = n_tracked_points
        self.block_size = block_size
        self.roi_windows = roi_windows

    def transform_data(self):
        trials_abs_w_ave, w, rn, cum_rn, w_fft, *_ = \
            et_mean(epochs=np.abs(self.input_node.data),
                    block_size=max(self.block_size, 5),
                    samples_distance=int(max(self.input_node.data.shape[0] // self.n_tracked_points, 10)),
                    roi_windows=self.roi_windows,
                    weighted=self.weighted_average
                    )
        w_ave, w, rn, cum_rn, w_fft, *_ = \
            et_mean(epochs=self.input_node.data,
                    block_size=max(self.block_size, 5),
                    samples_distance=int(max(self.input_node.data.shape[0] // self.n_tracked_points, 10)),
                    roi_windows=self.roi_windows,
                    weighted=self.weighted_average
                    )
        self.output_node.data = trials_abs_w_ave - np.abs(w_ave)
        self.output_node.rn = rn
        self.output_node.cum_rn = cum_rn
        self.output_node.snr = None
        self.output_node.cum_snr = None
        self.output_node.s_var = None


class AverageTimeFrequencyResponse(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 time_window: u.Quantity = 0.3 * u.s,
                 sample_interval: u.Quantity = 0.004 * u.s,
                 topographic_channels=np.array([]),
                 title: str = '',
                 plot_x_lim=None,
                 plot_y_lim=None,
                 times=np.array([]),
                 fig_format='.png',
                 fontsize=12,
                 user_naming_rule: str = '',
                 spec_thresh=4,
                 average_mode='magnitude',
                 **kwargs):
        super(AverageTimeFrequencyResponse, self).__init__(input_process=input_process, **kwargs)
        self.time_window = time_window
        self.sample_interval = sample_interval
        self.frequency = None
        self.time = None
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = times
        self.time_window = time_window
        self.sample_interval = sample_interval
        self.spec_thresh = spec_thresh
        self.average_mode = average_mode

    def transform_data(self):
        power, time, freqs = et_average_time_frequency_transformation(epochs=self.input_node.data,
                                                                      fs=self.input_node.fs,
                                                                      time_window=self.time_window,
                                                                      sample_interval=self.sample_interval,
                                                                      average_mode=self.average_mode
                                                                      )
        self.output_node.data = power
        self.output_node.x = time
        self.output_node.y = freqs
        self.output_node.domain = Domain.time_frequency


class PlotTimeFrequencyData(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 topographic_channels=np.array([]),
                 title='',
                 plot_x_lim=None,
                 plot_y_lim=None,
                 times=np.array([]),
                 fig_format='.png',
                 fontsize=12,
                 user_naming_rule='',
                 spec_thresh=4,
                 **kwargs):
        super(PlotTimeFrequencyData, self).__init__(input_process=input_process, **kwargs)
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = times
        self.spec_thresh = spec_thresh

    def transform_data(self):
        assert self.input_node.domain == Domain.time_frequency, 'input should be a time-frequency transformed data'

        eegpt.plot_eeg_time_frequency_power(ave_data=self.input_node,
                                            time=self.input_node.x,
                                            frequency=self.input_node.y,
                                            eeg_topographic_map_channels=self.topographic_channels,
                                            figure_dir_path=self.input_node.paths.figures_current_dir,
                                            figure_basename='{:}{:}'.format(
                                                self.input_process.name + '_',
                                                self.user_naming_rule),
                                            time_unit='s',
                                            amplitude_unit='uV',
                                            times=self.times,
                                            title=self.title,
                                            x_lim=self.plot_x_lim,
                                            y_lim=self.plot_y_lim,
                                            fig_format=self.fig_format,
                                            fontsize=self.fontsize,
                                            spec_thresh=self.spec_thresh
                                            )


class AverageSpectrogram(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 topographic_channels=np.array([]),
                 title='',
                 plot_x_lim=None,
                 plot_y_lim=None,
                 times=np.array([]),
                 fig_format='.png',
                 fontsize=12,
                 user_naming_rule='',
                 time_window=2.0,
                 sample_interval=0.004,
                 spec_thresh=4
                 ):
        super(AverageSpectrogram, self).__init__(input_process=input_process)
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = times
        self.time_window = time_window
        self.sample_interval = sample_interval
        self.spec_thresh = spec_thresh

    def transform_data(self):
        power, freqs = et_average_frequency_transformation(epochs=self.input_node.data,
                                                           fs=self.input_node.fs,
                                                           ave_mode='magnitude'
                                                           )
        self.output_node.data = power


class AverageFrequencyPower(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 topographic_channels=np.array([]),
                 title='',
                 plot_x_lim=None,
                 plot_y_lim=None,
                 times=np.array([]),
                 fig_format='.png',
                 fontsize=12,
                 user_naming_rule='',
                 **kwargs):
        super(AverageFrequencyPower, self).__init__(input_process=input_process, **kwargs)
        self.frequency = None
        self.time = None
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = times

    def transform_data(self):
        power, freqs = et_average_frequency_transformation(epochs=self.input_node.data,
                                                           fs=self.input_node.fs,
                                                           ave_mode='magnitude'
                                                           )
        self.output_node.data = power



"""
.. _tut-acc-standard-pipeline-sim:

##################################
ACC Standard analysis (Simulated)
##################################

In this example we simulate an ACC response and detect the peaks using n standard pipeline.

.. contents:: Page contents
   :local:
   :depth: 2
"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.bootstrap.bootstrap import Bootstrap, BootstrapTarget
from peegy.processing.pipe.general import FilterData, ReSampling, AutoRemoveBadChannels, ReferenceData, RegressOutEOG, \
    BaselineCorrection
from peegy.processing.pipe.epochs import EpochData, AverageEpochs
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.statistics import HotellingT2Test
from peegy.processing.pipe.detection import PeakDetectionTimeDomain
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import numpy as np
import astropy.units as u

# %%
# Time windows
# ============================
# Time windows are used to define where to look for different peaks


tw = np.array([TimePeakWindow(ini_time=50 * u.ms, end_ref='N1', label='P1', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=80 * u.ms, end_time=200 * u.ms, label='N1', positive_peak=False,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_ref='N1', end_time=300 * u.ms, label='P2', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=50 * u.ms, end_time=150 * u.ms, label='gfp1', positive_peak=True,
                              target_channels=['GFP']),
               TimePeakWindow(ini_time=150 * u.ms, end_time=500 * u.ms, label='gfp2', positive_peak=True,
                              target_channels=['GFP'])])

# %%
# Peak-to-peak Measures
# ============================
# Peak-to-peak measures are defined based on the labels of the TimePeaks defined above.


pm = np.array([PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])

# %%
# Time regions of interest
# ============================
# TimeROI are defined as time regions where different measures will be performed, e.g. SNR measure or statistical
# measures


roi_windows = np.array(
    [TimeROI(ini_time=100.0 * u.ms, end_time=300.0 * u.ms, measure="snr", label="acc_snr"),
     TimeROI(ini_time=500.0 * u.ms, end_time=700.0 * u.ms, measure="snr", label="control")])

# %%
# Generate some data
# ===================
# First we generate some ACC data

n_channels=32
fs = 512.0 * u.Hz
template_waveform, _ = aep(fs=fs)
event_times = np.arange(0, 100.0, 1.0) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           layout_file_name='biosemi32.lay',
                           snr=0.1,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='acc_test',
                           include_eog_events=True,
                           )
reader.run()

# %%
# Start the pipeline
# ============================
# Now we proceed with our basic processing pipeline


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline.referenced)
pipeline['down_sampled'] = ReSampling(pipeline.channel_cleaned,
                                      new_sampling_rate=256. * u.Hz)
pipeline['eog_removed'] = RegressOutEOG(pipeline.down_sampled,
                                        ref_channel_labels=['EOG1'])
pipeline.run()
# %%
# Show EOG removal Output
# ------------------------------------

pipeline.eog_removed.plot(plot_input=True,
                          plot_output=True,
                          ch_to_plot=['CP1', 'CP5', 'P7'],
                          interactive=False)

# %%
# Continue with the pipeline
# ------------------------------------

pipeline['time_filtered_data'] = FilterData(pipeline.eog_removed,
                                            high_pass=2.0 * u.Hz,
                                            low_pass=30.0 * u.Hz)

pipeline.run()

# %%
# Bootstrapping
# ------------------------------------
# We generate a new pipeline indicating what would be bootstrapped
pipeline_bootstrap = PipePool()

pipeline_bootstrap['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                              event_code=1.0,
                                              post_stimulus_interval=1 * u.s)
pipeline_bootstrap['time_epochs_dss'] = CreateAndApplySpatialFilter(input_process=pipeline_bootstrap['time_epochs'],
                                                                    plot_projections=False,
                                                                    components_to_plot=None,
                                                                    plot_power=False)
pipeline_bootstrap['time_epochs_with_baseline_corrected'] = BaselineCorrection(
    pipeline_bootstrap['time_epochs_dss'],
    ini_time=0 * u.ms,
    end_time=20 * u.ms)
pipeline_bootstrap['ht2'] = HotellingT2Test(pipeline_bootstrap['time_epochs_with_baseline_corrected'],
                                            roi_windows=roi_windows)
pipeline_bootstrap['ftest'] = AverageEpochs(pipeline_bootstrap['time_epochs_with_baseline_corrected'],
                                            roi_windows=roi_windows,
                                            weighted_average=False)

# bootstrap the pipe
pipeline['bootstrap'] = Bootstrap(at_each_bootstrap_do=pipeline_bootstrap,
                                  n_bootstraps=100,
                                  event_code=1.0,
                                  bootstrap_targets=[
                                      BootstrapTarget(
                                          table_name='hotelling_t2_time',
                                          group_by=['channel', 'label', 'ini_time', 'end_time'],
                                          target_values=['f']),
                                      BootstrapTarget(
                                          table_name='f_test_time',
                                          group_by=['channel', 'label', 'ini_time', 'end_time'],
                                          target_values=['f'])])
pipeline.run()

# now we save our data to a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ACC'}
database_path = reader.input_node.paths.file_directory + os.sep + 'acc_test_bootstrap_data.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['bootstrap']]
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



import unittest
import peegy.processing.tools.detection.time_domain_tools as td
from peegy.processing.pipe.definitions import DataNode
import numpy as np

__author__ = 'bwilliges-uoc'


class TestPeakDetectionMethods(unittest.TestCase):

    def test_peak(self):
        # create simple peak data
        fs = 44100.0
        nsamples = int(np.round(1 * fs))
        data = np.zeros((nsamples, 1))
        p1peak = int(np.round(0.1 * fs))  # P1 peak at 100 ms
        n1peak = int(np.round(0.15 * fs))  # N1 peak at 150 ms
        p2peak = int(np.round(0.2 * fs))  # P2 peak at 200 ms
        data[p1peak] = 1  # P1 peak
        data[n1peak] = -1  # N1 peak
        data[p2peak] = 1  # P2 peak
        dataobject = DataNode(fs=fs, data=data, rn=np.zeros(1)+1, snr=np.zeros(1), s_var=np.zeros(1)+1)
        tw = np.array([td.TimePeakWindow(ini_time=50e-3, end_time=150e-3, label='P1', positive_peak=True),
                       td.TimePeakWindow(ini_time=100e-3, end_time=200e-3, label='N1', positive_peak=False),
                       td.TimePeakWindow(ini_time=110e-3, end_time=300e-3, label='P2', positive_peak=True)])
        pm = np.array([td.PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])
        peak_containers, amplitudes = td.detect_peaks_and_amplitudes(dataobject,
                                                                     time_peak_windows=tw,
                                                                     peak_to_peak_measures=pm)
        a = peak_containers[0].get_peaks()
        self.assertEqual([a[0].idx, a[1].idx, a[2].idx], [p1peak, n1peak, p2peak])
        # self.assertEqual(amplitudes, 2)
        # later test case with reference window
        twref = np.array([td.TimePeakWindow(ini_time=50e-3, end_ref='N1', label='P1', positive_peak=True),
                          td.TimePeakWindow(ini_time=100e-3, end_time=200e-3, label='N1', positive_peak=False),
                          td.TimePeakWindow(ini_ref='N1', end_time=300e-3, label='P2', positive_peak=True)])
        pm = np.array([td.PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])
        peak_containers, amplitudes = td.detect_peaks_and_amplitudes(dataobject, time_peak_windows=twref,
                                                                     peak_to_peak_measures=pm)
        a = peak_containers[0].get_peaks()
        self.assertEqual([a[1].idx, a[0].idx, a[2].idx], [p1peak, n1peak, p2peak])
        self.assertEqual([a[0].idx, a[1].idx, a[2].idx], [p1peak, n1peak, p2peak])
        # self.assertEqual(amplitudes, 2)


if __name__ == '__main__':
    unittest.main()



import peegy.io.edf_bdf_reader as ebr
import peegy.io.generic_csv_reader as cvsr
import os
import abc
import numpy as np
import astropy.units as u


class EEGData(object, metaclass=abc.ABCMeta):
    def __init__(self, file_name=''):
        self.file_name = file_name
        self.fs = None
        self.n_channels = None
        self.default_layout = None
        self.event_channel_label = None
        self.data_unit = None

    def get_data(self, channels: np.array = np.array([]), ini_time=0, end_time: float = None):
        print('need to define')

    def get_events(self):
        print('need to define')


class _BDFEDFDataReader(EEGData):
    def __init__(self, file_name='',
                 event_channel_label: ebr.DeviceEventChannel = ebr.DeviceEventChannel.auto,
                 data_unit: u.Unit = u.uV):
        """
        This class handles reading of edf and bdf files
        :param file_name: path to file to be read
        :param event_channel_label: string indicating the label of the channel containing all events.
        This is usually the 'Status' channel for bdf files and 'EDF Annotations' for EDF files, however, some
        devices provide other labels, e.g. 'BDF Annotations', therefore we leave this open
        :param data_unit: defines what data will be read from the files. This allows for filtering data with mixing
        units (uV, Celciusm, uS, etcetera).
        """
        super(_BDFEDFDataReader, self).__init__(file_name=file_name)
        self._header = ebr.read_edf_bdf_header(self.file_name)
        self.n_channels = self._header['n_channels']
        self.fs = self._header['fs'][0]
        self.default_layout = None
        self.data_unit = data_unit
        _, file_extension = os.path.splitext(file_name)
        if file_extension == '.bdf':
            if event_channel_label == ebr.DeviceEventChannel.auto or event_channel_label is None:
                for ch in self._header['channels']:
                    if ch.label == ebr.DeviceEventChannel.bdf_event_channel:
                        self.event_channel_label = ebr.DeviceEventChannel.bdf_event_channel
                        break
                    if ch.label == ebr.DeviceEventChannel.bdf_event_annotations:
                        self.event_channel_label = ebr.DeviceEventChannel.bdf_event_annotations
        if file_extension == '.edf':
            self.event_channel_label = ebr.DeviceEventChannel.edf_event_channel
        # we override the default event_channel_label if another is provided by user
        if event_channel_label is not None and event_channel_label != ebr.DeviceEventChannel.auto:
            self.event_channel_label = event_channel_label

    def get_data(self, channels_idx: np.array = np.array([]),
                 ini_time: u.quantity.Quantity = 0 * u.s,
                 end_time: u.quantity.Quantity = None):
        data, events, units, annotations, valid_idx = ebr.get_data(header=self._header,
                                                                   channels_idx=channels_idx,
                                                                   ini_time=ini_time,
                                                                   end_time=end_time,
                                                                   event_channel_label=self.event_channel_label,
                                                                   data_unit=self.data_unit)
        self.default_layout = self._header['channels'][valid_idx]
        # here the event channel is removed from the layout
        _status_idx = [_i for _i, _ch in enumerate(self.default_layout) if
                       self.default_layout[_i].label == self.event_channel_label]
        if _status_idx:
            self.default_layout = np.delete(self.default_layout, _status_idx)
        return data, events, units, annotations

    def get_events(self):
        _events, _ = ebr.get_event_channel(header=self._header,
                                           event_channel_label=self.event_channel_label)
        return _events


class _CSVDataReader(EEGData):
    def __init__(self, file_name='',
                 fs_col_name: str = None,
                 gain_col_name: str = None,
                 fs_unit: u.quantity.Quantity = u.Hz,
                 gain_unit: u.quantity.Quantity = u.uV,
                 gain_inverted: bool = False
                 ):
        super(_CSVDataReader, self).__init__(file_name=file_name)
        self._header = cvsr.read_header(self.file_name,
                                        fs_col_name=fs_col_name,
                                        gain_col_name=gain_col_name,
                                        fs_unit=fs_unit,
                                        gain_unit=gain_unit,
                                        gain_inverted=gain_inverted
                                        )
        self.n_channels = self._header['n_channels']
        self.fs = self._header['fs']
        self.default_layout = None
        _, file_extension = os.path.splitext(file_name)
        if file_extension == '.txt':
            self.event_channel_label = "Events"

    def get_data(self, channels_idx: np.array = np.array([]), ini_time=0, end_time: float = None) -> (np.array,
                                                                                                      np.array,
                                                                                                      u,
                                                                                                      tuple):
        annotations = ()
        data, events, units = cvsr.read_channel(
            header=self._header,
            channels_idx=channels_idx,
            ini_time=ini_time,
            end_time=end_time)
        if channels_idx.size:
            self.default_layout = self._header['channels'][channels_idx]
        else:
            self.default_layout = self._header['channels']

        # here the status channel is removed from the layout
        _status_idx = [_i for _i, _ch in enumerate(self.default_layout) if
                       self.default_layout[_i].label == self.event_channel_label]
        if _status_idx:
            self.default_layout = np.delete(self.default_layout, _status_idx)
        return data, events, units, annotations

    def get_events(self):
        _, events, _ = cvsr.read_channel(
            header=self._header)
        return events


def eeg_reader(file_name='',
               fs_col_name: str = None,
               gain_col_name: str = None,
               fs_unit: u.quantity.Quantity = u.Hz,
               gain_unit: u.quantity.Quantity = u.uV,
               gain_inverted: bool = False,
               event_channel_label: ebr.DeviceEventChannel = None,
               data_unit: u.Unit = u.uV
               ):
    _, file_extension = os.path.splitext(file_name)
    if file_extension in ['.bdf', '.edf']:
        return _BDFEDFDataReader(file_name=file_name,
                                 event_channel_label=event_channel_label,
                                 data_unit=data_unit)
    if file_extension in ['.txt', '.cvs']:
        return _CSVDataReader(file_name=file_name,
                              fs_col_name=fs_col_name,
                              gain_col_name=gain_col_name,
                              fs_unit=fs_unit,
                              gain_unit=gain_unit,
                              gain_inverted=gain_inverted)



from tqdm.auto import tqdm
from joblib import Parallel


class ParallelTqdm(Parallel):
    def __init__(self, use_tqdm=True, total=None, desc='', *args, **kwargs):
        self._use_tqdm = use_tqdm
        self._total = total
        self._desc = desc
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        with tqdm(disable=not self._use_tqdm,
                  total=self._total,
                  desc=self._desc) as self._pbar:
            return Parallel.__call__(self, *args, **kwargs)

    def print_progress(self):
        if self._total is None:
            self._pbar.total = self.n_dispatched_tasks
        self._pbar.n = self.n_completed_tasks
        self._pbar.refresh()



import os
from os.path import sep
import logging
import numpy as np
from peegy.definitions.channel_definitions import ChannelItem
import matplotlib.pyplot as plt
__author__ = 'jundurraga'
log = logging.getLogger()


class Layout(object):
    def __init__(self, file_name='biosemi64.lay'):
        self.layout = self.get_layout(file_name=file_name)

    def get_layout(self, file_name='biosemi64.lay'):
        """
        Reads in a .lay Layout file, as specified by the variable file_name.
        The function will recognize all layouts inside of the "Layouts" subdirectory of peegy-Python.
        To ease working with subject specific Layout files, it is also possible to include the full path in the
        filename.
        :file_name: string containing the Layout file name (can contain a full path).
        """
        _path, file_name = os.path.split(file_name)
        if not _path:
            _path = os.path.dirname(os.path.realpath(__file__))
        return self.read_layout(_path + sep + file_name)

    @staticmethod
    def read_layout(file_path=None):
        out = np.array([])
        if file_path is not None:
            with open(file_path, 'r') as f:
                while True:
                    _data = f.readline()
                    if not _data:
                        break
                    lay = _data.rstrip().split('\t')
                    out = np.append(out,
                                    ChannelItem(idx=int(lay[0]) - 1,
                                                x=float(lay[1]),
                                                y=float(lay[2]),
                                                w=float(lay[3]),
                                                h=float(lay[4]),
                                                label=lay[5]))
        return out

    def get_labels(self):
        return np.array([_ch.label for _ch in self.layout])

    def get_index(self):
        return np.array([_ch.idx for _ch in self.layout])

    def get_item(self, idx=None):
        _idx = max(0, min(idx, self.layout.size))
        return self.layout[_idx]

    @staticmethod
    def change_layout_idem(layout={}, labels_to_replace=[['from', 'to']], channel_mapping=[]):
        for _pair in labels_to_replace:
            _l_from = _pair[0]
            _l_to = _pair[1]
            _pos_ch_from = [(_pos, layout['channel'][_pos]) for _pos, _ch in enumerate(layout['label']) if
                            _ch == _l_from]
            _pos_ch_to = [(_pos, layout['channel'][_pos]) for _pos, _ch in enumerate(layout['label']) if
                          _ch == _l_to]

            if _pos_ch_from:
                number_in_map_from = [(_pos, _ch['number']) for _pos, _ch in enumerate(channel_mapping) if
                                      _ch['number'] == _pos_ch_from[0][1]]
            else:
                number_in_map_from = [(_pos, _ch['number']) for _pos, _ch in enumerate(channel_mapping) if
                                      _ch['label'] == _l_from]

            if _pos_ch_to:
                number_in_map_to = [(_pos, _ch['number']) for _pos, _ch in enumerate(channel_mapping) if
                                    _ch['number'] == _pos_ch_to[0][1]]
            else:
                number_in_map_to = [(_pos, _ch['number']) for _pos, _ch in enumerate(channel_mapping) if
                                    _ch['label'] == _l_to]

            if number_in_map_from and number_in_map_to:
                ini_channel = number_in_map_from[0][1]
                end_channel = number_in_map_to[0][1]
                if _pos_ch_from:
                    _pos_from = _pos_ch_from[0][0]
                    layout['channel'][_pos_from] = end_channel
                else:
                    channel_mapping[number_in_map_from[0][0]]['number'] = end_channel

                text = 'remapping channel {:s} to {:s}'.format(_l_from, _l_to)
                logging.info('\n' + ''.join(text))
                print(text)

                if _pos_ch_to:
                    _pos_to = _pos_ch_to[0][0]
                    layout['channel'][_pos_to] = ini_channel
                else:
                    channel_mapping[number_in_map_to[0][0]]['number'] = ini_channel

                text = 'remapping channel {:s} to {:s}'.format(_l_to, _l_from)
                logging.info('\n' + ''.join(text))
                print(text)
            else:
                text = 'could not remap channel {:s} by {:s}, ' \
                       'one of the electrode pairs is not in the map'.format(_l_to, _l_from)
                logging.info('\n' + ''.join(text))
                print(text)

    def plot_layout(self):
        fig_out = plt.figure(constrained_layout=True)
        widths = [1.0]
        heights = [1.0]
        gs = fig_out.add_gridspec(ncols=1, nrows=1, width_ratios=widths,
                                  height_ratios=heights)
        inch = 2.54
        # fig_out.subplots_adjust(wspace=0.0, hspace=0.0)
        fig_out.set_size_inches(20.0 / inch, 20.0 / inch)
        ax = plt.subplot(gs[0, 0])
        for i, lay in enumerate(self.layout):
            ax.plot(lay.x, lay.y, 'o', color='b', markersize=1)
            ax.text(lay.x, lay.y, lay.label)
        return fig_out



from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.epochs_processing_tools import et_impulse_response_artifact_subtraction
from sklearn.linear_model import LinearRegression
import numpy as np
import os
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
from PyQt5.QtCore import QLibraryInfo
from tqdm import tqdm
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class RegressOutArtifact(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 event_code: float = None,
                 artifact_window: float = 0.01,
                 alternating_polarity: bool = False,
                 stimulus_waveform: np.array = None,
                 method: str = 'regression',
                 **kwargs):
        """
        This InputOutputProcess will remove an artifact which has the shape of the input stimulus_waveform
        :param input_process: InputOutputProcess Class
        :param event_code: event code where the stimulus waveform starts
        :param artifact_window: time from onset where we only expect artifacts (free of neural response)
        :param alternating_polarity: indicate whether the paradigm used an alternating polarity scheme. the
        stimulus_waveform will be alternating in polarity accord
        :param stimulus_waveform: single column numpy array with the target waveform artifact
        :param method: either 'regression' or 'xcorr'. A regression model or an scaled amplitude based on cross
        correlation will be used to estimate the amount of artifact
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RegressOutArtifact, self).__init__(input_process=input_process, **kwargs)
        self.event_code = event_code
        self.artifact_window = set_default_unit(artifact_window, u.s)
        self.alternating_polarity = alternating_polarity
        self.stimulus_waveform = set_default_unit(stimulus_waveform, u.uV)
        self.method = method

    def transform_data(self):
        data = self.input_node.data.copy()
        # if event code is given, convolve source with dirac train to generate entire signal
        subset_idx = np.arange(0, int(self.input_node.fs * self.artifact_window))
        _events_index = None
        if self.event_code is not None:
            _events_index = self.input_node.events.get_events_index(code=self.event_code, fs=self.input_node.fs)
            events = np.zeros(data.shape[0])
            events[_events_index] = 1
            if self.alternating_polarity:
                events[_events_index[1: -1: 2]] = -1
            #  convolve source with triggers
            source = filt_data(data=self.stimulus_waveform, b=events.flatten(), mode='full', onset_padding=False)
            source = source[0: data.shape[0]]
        else:
            source = self.stimulus_waveform[0: data.shape[0]]

        if self.method == 'regression':
            for _idx in np.arange(data.shape[1]):
                regression = LinearRegression()
                regression.fit(source[subset_idx], data[subset_idx, _idx])
                data[:, _idx] -= regression.predict(source) * source.unit
        if self.method == 'xcorr':
            if data.ndim == 2:
                transmission_index = data[subset_idx, :].T.dot(source[subset_idx]) / np.expand_dims(
                    np.sum(np.square(source[subset_idx])), axis=0)
                data -= np.tile(np.atleast_2d(source), (1, data.shape[1])) * transmission_index.T
            if data.ndim == 3:
                ave_data = np.mean(data, 2)
                transmission_index = ave_data[subset_idx, :].T.dot(source[subset_idx]) / np.expand_dims(
                    np.sum(np.square(source[subset_idx])), axis=0)
                data -= np.tile(np.atleast_3d(source * transmission_index.T), (1, 1, data.shape[2]))
        if self.method == 'xcorr_by_event' and _events_index is not None:
            reconstructed = np.zeros(data.shape)
            # we recreate a peak template
            _template = self.stimulus_waveform
            for _idx in tqdm(_events_index, desc='Artefact reconstruction'):
                _current_template = _template
                _ini = _idx
                _end = _idx + _template.shape[0]
                if _end > data.shape[0]:
                    _current_template = _template[0:-(_end - data.shape[0]), :]
                original = data[_ini: _end, :]
                transmission_index = original[subset_idx, :].T.dot(_current_template[subset_idx]) / np.expand_dims(
                    np.sum(np.square(_current_template[subset_idx])), axis=0)
                reconstructed[_ini: _end, :] = np.repeat(
                    _current_template,
                    data.shape[1],
                    axis=1) * transmission_index.T
            data -= reconstructed

        self.output_node.data = data


class RegressOutArtifactIR(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 stimulus_waveform: np.array = None,
                 ir_length: float = 0.01,
                 regularization_factor: float = 1,
                 **kwargs):
        """
        This InputOutputProcess will remove an artifact which has the shape of the input stimulus_waveform.
        Here, artifact is estimated from the impulse response. The latter is estimated by averaging individual
        impulse responses across epochs.
        :param input_process: InputOutputProcess Class
        stimulus_waveform will be alternating in polarity accord
        :param stimulus_waveform: single column numpy array with the target waveform artifact
        :param ir_length: estimated length in seconds of impulse response
        :param regularization_factor: This value is used to regularize the deconvolution based on Tikhonov
        regularization allow to estimate the transfer function when the input signal is sparse in frequency components
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RegressOutArtifactIR, self).__init__(input_process=input_process, **kwargs)
        self.stimulus_waveform = stimulus_waveform
        self.ir_length = ir_length
        self.regularization_factor = regularization_factor
        self.recovered_artifact = None

    def transform_data(self):
        data = self.input_node.data.copy()
        # compute impulse response
        clean_data, _recovered_artifact = et_impulse_response_artifact_subtraction(
            data=data,
            stimulus_waveform=self.stimulus_waveform,
            ir_length=int(self.ir_length * self.input_node.fs),
            regularization_factor=self.regularization_factor)
        self.recovered_artifact = _recovered_artifact
        self.output_node.data = clean_data



import unittest
import numpy as np
import peegy.processing.tools.eeg_epoch_operators as et


class DSSTests(unittest.TestCase):
    def test_et_demean(self):
        x = np.ones((100, 20, 30)) * np.arange(20)[:, None]
        dm_x = et.et_demean(x)
        self.assertAlmostEqual(np.sum(np.mean(dm_x, axis=0)), 0, delta=1e-12, msg="Should be 0")

    def test_et_demean_weights_1(self):
        x = np.random.randn(100, 20, 300) * np.arange(20)[:, None]
        w = np.ones((100, 1, 300))
        dm_x = et.et_demean(x, w)
        self.assertAlmostEqual(np.sum(np.mean(dm_x, axis=0)), 0, delta=1e-12, msg="Should be 0")

    def test_et_demean_weights_2(self):
        x = np.ones((100, 20, 300)) * np.arange(20)[:, None]
        w = 2 * np.ones((100, 20, 300))
        dm_x = et.et_demean(x, w)
        self.assertAlmostEqual(np.sum(np.mean(dm_x, axis=0)), 0, delta=1e-12, msg="Should be 0")

    def test_et_demean_weights_3(self):
        x = np.ones((100, 20, 300))
        w = np.arange(20)
        dm_x = et.et_demean(x, w)
        self.assertAlmostEqual(np.sum(np.mean(dm_x, axis=0)), 0, delta=1e-12, msg="Should be 0")

    def test_et_weighted_cov_1(self):
        x = np.ones((100, 20, 30)) * np.arange(20)[:, None]
        c1, t1 = et.et_weighted_covariance(x)
        c2, t2 = et.et_covariance(x)
        self.assertTrue(np.all(c1 == c2), msg="Should be True")

        w = np.ones(x.shape) * 2
        c1, t1 = et.et_weighted_covariance(x, w)
        c2, t2 = et.et_covariance(x)
        self.assertTrue(np.all(c1 / t1 == c2/t2), msg="Should be True")

    def test_et_weighted_cov_2(self):
        x = np.ones((100, 20, 30)) * np.arange(20)[:, None]
        w = np.ones(x.shape)
        w = w * np.arange(1, x.shape[2] + 1)
        c1, t1 = et.et_weighted_covariance(x, w)
        c2, t2 = et.et_covariance(x)
        self.assertTrue(np.all(c1 / t1 == c2/t2), msg="Should be True")


if __name__ == '__main__':
    unittest.main()



"""
.. _tut-acc-sim:

Test spatial filtering with evoked response (Simulation)
==============================

This example simulates an evoked responses with non-stationary noise.
We illustrate how weighted averaging and DSS can lead to much better results in terms of recovering the data.
This example follows that by Alain de Cheveigné (http://audition.ens.fr/adc/NoiseTools/) but it has been extended
to include non-stationary noise and weighted average.
.. contents:: Page contents
   :local:
   :depth: 2
"""
import matplotlib.pyplot as plt
from peegy.processing.tools.eeg_epoch_operators import et_mmat
import numpy as np
from scipy.stats import norm
from astropy import units as u
# Enable below for interactive backend
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')
__author__ = 'jundurraga'

# create synthetic data
nsamples = 500
nchans = 1
ntrials = 100
noise_dim = 20  # dimensionality of noise
freq = 100
fs = 1000

source = np.sin(2 * np.pi * np.arange(nsamples) / fs * freq)
np.random.seed(2)
mix_m = np.ones((nchans, 1))
s = source * mix_m
s = np.tile(np.expand_dims(s.T, axis=2), (1, 1, ntrials))
SNR = 5

noise_unmixed = norm.ppf(np.random.rand(noise_dim * ntrials * nsamples))
noise_unmixed = np.reshape(noise_unmixed, [nsamples, noise_dim, ntrials], order='F')

noise_mix = norm.ppf(np.random.rand(nchans * noise_dim))
noise_mix = np.reshape(noise_mix, [noise_dim, nchans], order='F')
noise = et_mmat(noise_unmixed, noise_mix)

scaled_source = s * u.uV / np.std(s)
scaled_noise = (1 / SNR) * np.std(s) * noise * u.uV / np.std(noise)

# mix data
data = scaled_noise + scaled_source
frequencies = freq * (1 + np.arange(1))

yfft = np.fft.fft(data, axis=0).value
freqs = np.arange(0, data.shape[0]) * fs / data.shape[0]
k = np.argmin(np.abs(freqs - freq))
yfft_bin = yfft[k, ...]

r = np.abs(yfft_bin)
ang = np.angle(yfft_bin)
mean_yfft_bin = np.abs(np.mean(yfft, axis=2))

plt.figure()
plt.plot(r * np.cos(ang), r * np.sin(ang), marker='o', color='b', linestyle="None")

noise_bin = yfft_bin - np.mean(yfft_bin, axis=1, keepdims=True)
mean_yfft_noise_bin = np.abs(np.mean(noise_bin, axis=1))
# noise_bin_power = np.sqrt(np.mean(np.abs(noise_bin) ** 2.0, axis=1) / noise_bin.shape[1])
noise_bin_power = np.var(noise_bin, axis=1, ddof=1) / noise_bin.shape[1]

# actual noise
yfft_noise = np.fft.fft(scaled_noise, axis=0).value
yfft_noise_bin = yfft_noise[k, ...]
yfft_noise_mean = np.mean(yfft_noise_bin, axis=1, keepdims=True)
yfft_noise_mean_power = np.abs(yfft_noise_mean) ** 2.0

noise_across_bins = np.sum(np.abs(np.mean(yfft_noise, axis=2)) ** 2, axis=0) / yfft_noise.shape[0]


plt.figure()
plt.plot(yfft_noise_mean_power)
plt.plot(noise_bin_power)
plt.show()

plt.figure()
plt.plot(freqs, mean_yfft_bin ** 2)
[plt.axhline(y=np.mean(_nb), color='orange') for _nb in noise_bin_power]
[plt.axhline(y=np.mean(_nb), color='red') for _nb in yfft_noise_mean_power]
plt.axhline(y=noise_across_bins, color='blue')
plt.show()


plt.figure()
plt.plot(freqs, np.abs(np.mean(yfft_noise, axis=2)) ** 2)
[plt.axhline(y=np.mean(_nb), color='orange') for _nb in noise_bin_power]
[plt.axhline(y=np.mean(_nb), color='red') for _nb in yfft_noise_mean_power]
plt.show()


####
r_n = np.abs(noise_bin)
ang_n = np.angle(noise_bin)
plt.plot(r_n * np.cos(ang_n), r_n * np.sin(ang_n), marker='o', color='orange', linestyle="None")
plt.plot(r_n * np.cos(ang_n), r_n * np.sin(ang_n), marker='o', color='orange', linestyle="None")
